@@ -1,7 +1,7 @@
-/* $Id: acl.c,v 1.97 2010/04/12 12:04:41 manu Exp $ */
+/* $Id: acl.c,v 1.111 2013/09/01 04:59:42 manu Exp $ */
 
 /*
- * Copyright (c) 2004-2007 Emmanuel Dreyfus
+ * Copyright (c) 2004-2012 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,11 +34,11 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: acl.c,v 1.97 2010/04/12 12:04:41 manu Exp $");
+__RCSID("$Id: acl.c,v 1.111 2013/09/01 04:59:42 manu Exp $");
 #endif
 #endif
 
-#ifdef HAVE_OLD_QUEUE_H
+#if defined(HAVE_OLD_QUEUE_H) || !defined(HAVE_SYS_QUEUE_H)
 #include "queue.h"
 #else 
 #include <sys/queue.h>
@@ -79,9 +79,7 @@ __RCSID("$Id: acl.c,v 1.97 2010/04/12 12:04:41 manu Exp $");
 #ifdef USE_LDAP
 #include "ldapcheck.h"
 #endif
-#if defined(USE_CURL) || defined(USE_LDAP)
 #include "prop.h"
-#endif
 #ifdef USE_GEOIP
 #include "geoip.h"
 #endif
@@ -112,6 +110,8 @@ pthread_rwlock_t acl_lock;
 static struct acl_entry *gacl;
 int gneg;
 
+static int acl_actions(struct mlfi_priv *, acl_stage_t, struct acl_entry *,
+		       struct acl_param *, int);
 char *acl_print_netblock(acl_data_t *, char *, size_t);
 char *acl_print_string(acl_data_t *, char *, size_t);
 char *acl_print_regex(acl_data_t *, char *, size_t);
@@ -153,14 +153,24 @@ char *acl_print_urlcheck(acl_data_t *, char *, size_t);
 void acl_add_ldapcheck(acl_data_t *, void *);
 char *acl_print_ldapcheck(acl_data_t *, char *, size_t);
 #endif
-#if defined(USE_CURL) || defined(USE_LDAP)
 void acl_add_prop_string(acl_data_t *, void *);
+void acl_add_prop_pop(acl_data_t *, void *);
+void acl_add_prop_opnum(acl_data_t *, void *);
 void acl_add_prop_regex(acl_data_t *, void *);
 char *acl_print_prop_string(acl_data_t *, char *, size_t);
+char *acl_print_prop_pop(acl_data_t *, char *, size_t);
+char *acl_print_prop_opnum(acl_data_t *, char *, size_t);
 char *acl_print_prop_regex(acl_data_t *, char *, size_t);
 void acl_free_prop_string(acl_data_t *);
+void acl_free_prop_pop(acl_data_t *);
+void acl_free_prop_opnum(acl_data_t *);
 void acl_free_prop_regex(acl_data_t *);
+#ifdef USE_NSUPDATE
+char *acl_print_nsupdate(acl_data_t *, char *, size_t);
+void acl_add_nsupdate(acl_data_t *, void *);
 #endif
+int acl_log_string(acl_data_t *, acl_stage_t, 
+		   struct acl_param *, struct mlfi_priv *);
 
 struct acl_clause_rec acl_clause_rec[] = {
 	/* Temporary types for lists */
@@ -250,6 +260,10 @@ struct acl_clause_rec acl_clause_rec[] = {
 	  AT_LIST, AC_NONE, AC_NONE, EXF_BODY,
 	  acl_print_list, acl_add_list, 
 	  NULL, acl_list_filter },
+	{ AC_BODY_PROP, MULTIPLE_OK, AS_DATA, "body_prop", 
+	  AT_STRING, AC_BODY_LIST, AC_STRING, EXF_BODY,
+	  acl_print_string, acl_add_body_string, 
+	  acl_free_string, prop_body_validate },
 	{ AC_HEADER, MULTIPLE_OK, AS_DATA, "header", 
 	  AT_STRING, AC_HEADER_LIST, AC_STRING, EXF_HEADER,
 	  acl_print_string, acl_add_body_string, 
@@ -262,6 +276,10 @@ struct acl_clause_rec acl_clause_rec[] = {
 	  AT_LIST, AC_NONE, AC_NONE, EXF_HEADER,
 	  acl_print_list, acl_add_list, 
 	  NULL, acl_list_filter },
+	{ AC_HEADER_PROP, MULTIPLE_OK, AS_DATA, "header_prop", 
+	  AT_STRING, AC_HEADER_LIST, AC_STRING, EXF_HEADER,
+	  acl_print_string, acl_add_body_string, 
+	  acl_free_string, prop_header_validate },
 	{ AC_MACRO, MULTIPLE_OK, AS_ANY, "macro", 
 	  AT_MACRO, AC_MACRO_LIST, AC_STRING, EXF_MACRO,
 	  acl_print_macro, acl_add_macro,
@@ -306,16 +324,22 @@ struct acl_clause_rec acl_clause_rec[] = {
 	  acl_print_ldapcheck, acl_add_ldapcheck,
 	  NULL, ldapcheck_validate },
 #endif
-#if defined(USE_CURL) || defined(USE_LDAP)
-	{ AC_PROP, MULTIPLE_OK, AS_ANY, "prop", 
-	  AT_PROP, AC_NONE, AC_PROP, EXF_PROP,
+	{ AC_PROP_STR, MULTIPLE_OK, AS_ANY, "prop", 
+	  AT_PROP, AC_NONE, AC_PROP_STR, EXF_PROP,
 	  acl_print_prop_string, acl_add_prop_string,
 	  acl_free_prop_string, prop_string_validate },
+	{ AC_PROP_PROP, MULTIPLE_OK, AS_ANY, "propprop", 
+	  AT_PROP, AC_NONE, AC_PROP_PROP, EXF_PROP,
+	  acl_print_prop_pop, acl_add_prop_pop,
+	  acl_free_prop_pop,  prop_pop_validate },
+	{ AC_PROP_NUM, MULTIPLE_OK, AS_ANY, "prop_num", 
+	  AT_PROP, AC_NONE, AC_PROP_NUM, EXF_PROP,
+	  acl_print_prop_opnum, acl_add_prop_opnum,
+	  acl_free_prop_opnum, prop_lhsnum_validate },
 	{ AC_PROP_RE, MULTIPLE_OK, AS_ANY, "prop_re", 
 	  AT_PROP, AC_NONE, AC_PROP_RE, EXF_PROP,
 	  acl_print_prop_regex, acl_add_prop_regex,
 	  acl_free_prop_regex, prop_regex_validate },
-#endif
 	{ AC_AUTH, MULTIPLE_OK, AS_ANY, "auth", 
 	  AT_STRING, AC_AUTH_LIST, AC_STRING, EXF_AUTH,
 	  acl_print_string, acl_add_string,
@@ -357,10 +381,18 @@ struct acl_clause_rec acl_clause_rec[] = {
 	  AT_OPNUM, AC_NONE, AC_MSGSIZE, EXF_MSGSIZE,
 	  acl_print_opnum, acl_add_opnum,
 	  NULL, acl_msgsize_cmp },
+	{ AC_MSGSIZE_PROP, MULTIPLE_OK, AS_DATA, "msgsize_prop", 
+	  AT_OPNUM, AC_NONE, AC_STRING_PROP, EXF_MSGSIZE,
+	  prop_opnum_print, prop_opnum_add,
+	  prop_opnum_free, prop_rhsnum_validate },
 	{ AC_RCPTCOUNT, MULTIPLE_OK, AS_ANY, "rcptcount", 
 	  AT_OPNUM, AC_NONE, AC_RCPTCOUNT, EXF_RCPTCOUNT,
 	  acl_print_opnum, acl_add_opnum_body,
 	  NULL, acl_rcptcount_cmp },
+	{ AC_RCPTCOUNT_PROP, MULTIPLE_OK, AS_DATA, "rcptcount_prop", 
+	  AT_OPNUM, AC_NONE, AC_STRING_PROP, EXF_RCPTCOUNT,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_rhsnum_validate },
 	{ AC_CLOCKSPEC, MULTIPLE_OK, AS_ANY, "time",
 	  AT_CLOCKSPEC, AC_NONE, AC_CLOCKSPEC, EXF_CLOCKSPEC,
 	  print_clockspec, add_clockspec,
@@ -400,13 +432,75 @@ struct acl_clause_rec acl_clause_rec[] = {
 	{ AC_SASCORE, MULTIPLE_OK, AS_DATA, "spamd score",
 	  AT_OPNUM, AC_NONE, AC_NONE,  EXF_SA,
 	  acl_print_opnum, acl_add_opnum, NULL, spamd_score },
-#endif
+	{ AC_SASCORE_PROP, MULTIPLE_OK, AS_DATA, "sascore_prop", 
+	  AT_OPNUM, AC_NONE, AC_STRING_PROP, EXF_SA,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_rhsnum_validate },
+#endif /* USE_SPAMD */
 #ifdef HAVE_DATA_CALLBACK
 	{ AC_TARPIT, UNIQUE, AS_ANY, "tarpit",
 	  AT_TIME, AC_NONE, AC_NONE, EXF_TARPIT,
 	  acl_print_time, acl_add_time,
 	  NULL, acl_tarpit_filter },
 #endif
+#ifdef USE_NSUPDATE
+	{ AC_NSUPDATE, MULTIPLE_OK, AS_ANY, "nsupdate",
+	  AT_NSUPDATE, AC_NONE, AC_STRING, EXF_NONE,
+	  acl_print_nsupdate, acl_add_nsupdate,
+	  NULL, nsupdate_filter },
+#endif
+	{ AC_EQSET, MULTIPLE_OK, AS_ANY, "set", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_eqset_string },
+	{ AC_EQRSET, MULTIPLE_OK, AS_ANY, "rset", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_eqrset_string },
+	{ AC_INCSET, MULTIPLE_OK, AS_ANY, "incset", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_incset_string },
+	{ AC_INCRSET, MULTIPLE_OK, AS_ANY, "incrset", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_incrset_string },
+	{ AC_DECSET, MULTIPLE_OK, AS_ANY, "decset", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_decset_string },
+	{ AC_DECRSET, MULTIPLE_OK, AS_ANY, "decrset", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_decrset_string },
+	{ AC_EQSETPROP, MULTIPLE_OK, AS_ANY, "setprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_eqset_prop },
+	{ AC_EQRSETPROP, MULTIPLE_OK, AS_ANY, "rsetprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_eqrset_prop },
+	{ AC_INCSETPROP, MULTIPLE_OK, AS_ANY, "incsetprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_incset_prop },
+	{ AC_INCRSETPROP, MULTIPLE_OK, AS_ANY, "incrsetprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_incrset_prop },
+	{ AC_DECSETPROP, MULTIPLE_OK, AS_ANY, "decsetprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_decset_prop },
+	{ AC_DECRSETPROP, MULTIPLE_OK, AS_ANY, "decrsetprop", 
+	  AT_PROP, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_prop_string, acl_add_prop_string,
+	  acl_free_prop_string, prop_decrset_prop },
+	{ AC_LOG, MULTIPLE_OK, AS_ANY, "log", 
+	  AT_STRING, AC_NONE, AC_NONE, EXF_NONE,
+	  acl_print_string, acl_add_string,
+	  acl_free_string, acl_log_string },
 };
 
 struct {
@@ -543,15 +637,15 @@ acl_header_strstr(ad, stage, ap, priv)
 	struct acl_param *ap;
 	struct mlfi_priv *priv;
 {
-	struct header *h;
+	struct line *l;
 	 
 	if (stage != AS_DATA) {
 		mg_log(LOG_ERR, "header filter called at non DATA stage");
 		exit(EX_SOFTWARE);
 	}
 
-	TAILQ_FOREACH(h, &priv->priv_header, h_list)
-		if (strstr(h->h_line, ad->string) != NULL)
+	TAILQ_FOREACH(l, &priv->priv_header, l_list)
+		if (strstr(l->l_line, ad->string) != NULL)
 			return 1;
 	return 0;
 }
@@ -563,15 +657,15 @@ acl_body_strstr(ad, stage, ap, priv)
 	struct acl_param *ap;
 	struct mlfi_priv *priv;
 {
-	struct body *b;
+	struct line *l;
 	 
 	if (stage != AS_DATA) {
 		mg_log(LOG_ERR, "body filter called at non DATA stage");
 		exit(EX_SOFTWARE);
 	}
 
-	TAILQ_FOREACH(b, &priv->priv_body, b_list)
-		if (strstr(b->b_lines, ad->string) != NULL)
+	TAILQ_FOREACH(l, &priv->priv_body, l_list)
+		if (strstr(l->l_line, ad->string) != NULL)
 			return 1;
 
 	return 0;
@@ -793,15 +887,15 @@ acl_header_regexec(ad, stage, ap, priv)
 	struct acl_param *ap;
 	struct mlfi_priv *priv;
 {
-	struct header *h;
+	struct line *l;
 	 
 	if (stage != AS_DATA) {
 		mg_log(LOG_ERR, "header filter called at non DATA stage");
 		exit(EX_SOFTWARE);
 	}
 
-	TAILQ_FOREACH(h, &priv->priv_header, h_list)
-		if (myregexec(priv, ad, ap, h->h_line) == 0)
+	TAILQ_FOREACH(l, &priv->priv_header, l_list)
+		if (myregexec(priv, ad, ap, l->l_line) == 0)
 			return 1;
 	return 0;
 }
@@ -927,15 +1021,15 @@ acl_body_regexec(ad, stage, ap, priv)
 	struct acl_param *ap;
 	struct mlfi_priv *priv;
 {
-	struct body *b;
+	struct line *l;
 	 
 	if (stage != AS_DATA) {
 		mg_log(LOG_ERR, "body filter called at non DATA stage");
 		exit(EX_SOFTWARE);
 	}
 
-	TAILQ_FOREACH(b, &priv->priv_body, b_list)
-		if (myregexec(priv, ad, ap, b->b_lines) == 0)
+	TAILQ_FOREACH(l, &priv->priv_body, l_list)
+		if (myregexec(priv, ad, ap, l->l_line) == 0)
 			return 1;
 	return 0;
 }
@@ -1020,11 +1114,9 @@ acl_print_null(ad, buf, len)
 	return buf;
 }
 
-char *
-acl_print_opnum(ad, buf, len)
-	acl_data_t *ad;
-	char *buf;
-	size_t len;
+static char *
+acl_print_op(op)
+	enum operator op;
 {
 	struct {
 		enum operator op;
@@ -1038,20 +1130,24 @@ acl_print_opnum(ad, buf, len)
 		{ OP_LE, "<=" },
 	};
 	int i;
-	char *str = NULL;
 
 	for (i = 0; i < sizeof(op_to_str) / sizeof(*op_to_str); i++) {
-		if (op_to_str[i].op == ad->opnum.op) {
-			str = op_to_str[i].str;
-			break;
-		}
-	}
-	if (str == NULL) {
-		mg_log(LOG_ERR, "unexpected operator");
-		exit(EX_SOFTWARE);
+		if (op_to_str[i].op == op)
+			return op_to_str[i].str;
 	}
 
-	snprintf(buf, len, "%s %d", str, ad->opnum.num);
+	mg_log(LOG_ERR, "unexpected operator");
+	exit(EX_SOFTWARE);
+}
+
+
+char *
+acl_print_opnum(ad, buf, len)
+	acl_data_t *ad;
+	char *buf;
+	size_t len;
+{
+	snprintf(buf, len, "%s %d", acl_print_op(ad->opnum.op), ad->opnum.num);
 
 	return buf;
 }
@@ -1207,6 +1303,15 @@ acl_init_entry(void)
 	acl->a_autowhite = -1;
 	acl->a_tarpit = -1;
 	acl->a_tarpit_scope = -1;
+
+	/*
+	 * First gacl initialzation is done before conf is
+	 * valid. We therefore have to set to 0 and have
+	 * maxpeek global setting overwriting gacl->a_maxpeek
+	 * in acl_maxpeek_fixup()
+	 */
+        acl->a_maxpeek = (&conf != NULL) ? conf.c_maxpeek : 0;
+
 	TAILQ_INIT(&acl->a_clause);
 
 	return acl;
@@ -1252,6 +1357,10 @@ acl_free_entry(acl)
 		free(acl->a_report);
 	if (acl->a_addheader != NULL)
 		free(acl->a_addheader);
+	if (acl->a_addfooter != NULL)
+		free(acl->a_addfooter);
+	if (acl->a_subjtag != NULL)
+		free(acl->a_subjtag);
 	free(acl);
 
 	return;
@@ -1521,7 +1630,6 @@ acl_print_ldapcheck(ad, buf, len)
 }
 #endif
 
-#if defined(USE_CURL) || defined(USE_LDAP)
 void
 acl_add_prop_string(ad, data)
 	acl_data_t *ad;
@@ -1551,6 +1659,62 @@ acl_add_prop_string(ad, data)
 
 	return;
 }
+
+void
+acl_add_prop_pop(ad, data)
+	acl_data_t *ad;
+	void *data;
+{
+	struct acl_prop_pop *apop = (struct acl_prop_pop *)data;
+
+	if ((ad->apop = malloc(sizeof(*ad->apop))) == NULL) {
+		mg_log(LOG_ERR, "acl malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+
+	if ((ad->apop->apop_rhs = strdup(apop->apop_rhs)) == NULL) {
+		mg_log(LOG_ERR, "acl strdup failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+
+	ad->apop->apop_op = apop->apop_op;
+
+	if ((ad->apop->apop_lhs = strdup(apop->apop_lhs)) == NULL) {
+		mg_log(LOG_ERR, "acl strdup failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+}
+
+void
+acl_add_prop_opnum(ad, data)
+	acl_data_t *ad;
+	void *data;
+{
+	struct prop_data *upd;
+
+	upd = (struct prop_data *)data;
+
+	if ((ad->prop = malloc(sizeof(*ad->prop))) == NULL) {
+		mg_log(LOG_ERR, "acl malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+
+	if ((ad->prop->upd_name = strdup(upd->upd_name + 1)) == NULL) {
+		mg_log(LOG_ERR, "acl strdup failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+
+	if ((ad->prop->upd_data = malloc(sizeof(acl_data_t))) == NULL) {
+		mg_log(LOG_ERR, "acl malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+	memset(ad->prop->upd_data, 0, sizeof(acl_data_t));
+
+	acl_add_opnum((void *)ad->prop->upd_data, upd->upd_data);
+
+	return;
+}
+
 void 
 acl_add_prop_regex(ad, data)
 	acl_data_t *ad;
@@ -1595,6 +1759,35 @@ acl_print_prop_string(ad, buf, len)
 	return buf;
 }
 
+char *
+acl_print_prop_pop(ad, buf, len)
+	acl_data_t *ad;
+	char *buf;
+	size_t len;
+{
+	snprintf(buf, len, "$%s %s $%s ",
+		 ad->apop->apop_lhs,
+		 acl_print_op(ad->apop->apop_op),
+		 ad->apop->apop_rhs);
+
+	return buf;
+}
+
+char *
+acl_print_prop_opnum(ad, buf, len)
+	acl_data_t *ad;
+	char *buf;
+	size_t len;
+{
+	size_t written;
+
+	written = snprintf(buf, len, "$%s ", ad->prop->upd_name);
+	acl_print_opnum(ad->prop->upd_data, buf + written, len - written);
+
+	return buf;
+}
+
+
 char *
 acl_print_prop_regex(ad, buf, len)
 	acl_data_t *ad;
@@ -1620,6 +1813,29 @@ acl_free_prop_string(ad)
 
 	return;
 }
+
+void 
+acl_free_prop_opnum(ad)
+	acl_data_t *ad;
+{
+	free(ad->prop->upd_data);
+	free(ad->prop->upd_name);
+	free(ad->prop);
+
+	return;
+}
+
+void 
+acl_free_prop_pop(ad)
+	acl_data_t *ad;
+{
+	free(ad->apop->apop_rhs);
+	free(ad->apop->apop_lhs);
+	free(ad->apop);
+
+	return;
+}
+
 void 
 acl_free_prop_regex(ad)
 	acl_data_t *ad;
@@ -1632,6 +1848,26 @@ acl_free_prop_regex(ad)
 	return;
 }
 
+#ifdef USE_NSUPDATE
+void
+acl_add_nsupdate(ad, data)
+	acl_data_t *ad;
+	void *data;
+{
+	ad->nsupdate = (struct nsupdate_entry *)data;
+}
+
+char *
+acl_print_nsupdate(ad, buf, len)
+	acl_data_t *ad;
+	char *buf;
+	size_t len;
+{
+	(void)snprintf(buf, len, "$%s ", ad->nsupdate->nse_name);
+
+	return buf;
+}
+
 #endif
 
 void
@@ -1808,6 +2044,9 @@ acl_register_entry_first(acl_stage, acl_type)/* acllist must be write-locked */
 		case A_BLACKLIST:
 			mg_log(LOG_DEBUG, "register acl first BLACKLIST");
 			break;
+		case A_CONTINUE:
+			mg_log(LOG_DEBUG, "register acl first CONTINUE");
+			break;
 		default:
 			mg_log(LOG_ERR, "unexpected acl_type %d", acl_type);
 			exit(EX_SOFTWARE);
@@ -1888,6 +2127,9 @@ acl_register_entry_last(acl_stage, acl_type)/* acllist must be write-locked */
 		case A_BLACKLIST:
 			mg_log(LOG_DEBUG, "register acl last BLACKLIST");
 			break;
+		case A_CONTINUE:
+			mg_log(LOG_DEBUG, "register acl last CONTINUE");
+			break;
 		default:
 			mg_log(LOG_ERR, "unexpected acl_type %d", acl_type);
 			exit(EX_SOFTWARE);
@@ -1898,6 +2140,171 @@ acl_register_entry_last(acl_stage, acl_type)/* acllist must be write-locked */
 	return acl;
 }
 
+static int
+acl_actions(priv, stage, acl, ap, retval)
+	struct mlfi_priv *priv;
+	acl_stage_t stage;
+	struct acl_entry *acl;
+	struct acl_param *ap;
+	int retval;
+{
+	struct sockaddr *sa;
+	socklen_t salen;
+	char addrstr[IPADDRSTRLEN];
+
+	sa = SA(&priv->priv_addr);
+	salen = priv->priv_addrlen;
+
+	if (retval == 0)
+		retval = EXF_DEFAULT;
+	switch (ap->ap_type) {
+	case A_GREYLIST:
+		retval |= EXF_GREYLIST;
+		break;
+	case A_WHITELIST:
+		retval |= EXF_WHITELIST;
+		break;
+	case A_BLACKLIST:
+		retval |= EXF_BLACKLIST;
+		break;
+	case A_CONTINUE:
+		break;
+	default:
+		mg_log(LOG_ERR, "corrupted acl list");
+		exit(EX_SOFTWARE);
+		break;
+	}
+
+	priv->priv_sr.sr_acl_line = acl->a_line;
+
+	priv->priv_sr.sr_delay =
+	    (ap->ap_delay != -1) ? ap->ap_delay : conf.c_delay;
+	priv->priv_sr.sr_autowhite =
+	    (ap->ap_autowhite != -1) ? 
+	    ap->ap_autowhite : conf.c_autowhite_validity;
+	priv->priv_sr.sr_tarpit =
+	    (ap->ap_tarpit != -1) ? ap->ap_tarpit : conf.c_tarpit;
+	priv->priv_sr.sr_tarpit_scope =
+	    (ap->ap_tarpit_scope != -1) ?
+	    ap->ap_tarpit_scope : conf.c_tarpit_scope;
+	if (ap->ap_tarpitted > priv->priv_max_tarpitted)
+		priv->priv_max_tarpitted = ap->ap_tarpitted;
+	if (ap->ap_tarpitted > priv->priv_total_tarpitted)
+		priv->priv_total_tarpitted = ap->ap_tarpitted;
+
+	if (ap->ap_id) {
+		priv->priv_sr.sr_acl_id = strdup(ap->ap_id);
+		if (priv->priv_sr.sr_acl_id == NULL) { 
+			mg_log(LOG_ERR, "strdup failed: %s", 
+			    strerror(errno));
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_code) {
+		priv->priv_sr.sr_code = strdup(ap->ap_code);
+		if (priv->priv_sr.sr_code == NULL) { 
+			mg_log(LOG_ERR, "strdup failed: %s", 
+			    strerror(errno));
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_ecode) {
+		priv->priv_sr.sr_ecode = strdup(ap->ap_ecode);
+		if (priv->priv_sr.sr_ecode == NULL) {
+			mg_log(LOG_ERR, "strdup failed: %s", 
+			    strerror(errno));
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_msg) {
+		priv->priv_sr.sr_msg = strdup(ap->ap_msg);
+		if (priv->priv_sr.sr_msg == NULL) {
+			mg_log(LOG_ERR, "strdup failed");
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_report) {
+		priv->priv_sr.sr_report = strdup(ap->ap_report);
+		if (priv->priv_sr.sr_report == NULL) {
+			mg_log(LOG_ERR, "strdup failed");
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_addheader) {
+		priv->priv_sr.sr_addheader = strdup(ap->ap_addheader);
+		if (priv->priv_sr.sr_addheader == NULL) {
+			mg_log(LOG_ERR, "strdup failed");
+			exit(EX_OSERR);
+		}
+	}
+	if (ap->ap_addfooter) {
+		priv->priv_sr.sr_addfooter = strdup(ap->ap_addfooter);
+		if (priv->priv_sr.sr_addfooter == NULL) {
+			mg_log(LOG_ERR, "strdup failed");
+			exit(EX_OSERR);
+		}
+		(void)fstring_escape(priv->priv_sr.sr_addfooter);
+	}
+	if (ap->ap_subjtag) {
+		priv->priv_sr.sr_subjtag = strdup(ap->ap_subjtag);
+		if (priv->priv_sr.sr_subjtag == NULL) {
+			mg_log(LOG_ERR, "strdup failed");
+			exit(EX_OSERR);
+		}
+	}
+
+	if (stage == AS_RCPT)
+		priv->priv_maxpeek = ap->ap_maxpeek;
+
+	priv->priv_sr.sr_nmatch = ap->ap_nmatch;
+	priv->priv_sr.sr_pmatch = ap->ap_pmatch;
+		
+	/* Free temporary memory if nescessary */
+	if (ap->ap_flags & A_FREE_ID)
+		free(ap->ap_id);
+	if (ap->ap_flags & A_FREE_CODE)
+		free(ap->ap_code);
+	if (ap->ap_flags & A_FREE_ECODE)
+		free(ap->ap_ecode);
+	if (ap->ap_flags & A_FREE_MSG)
+		free(ap->ap_msg);
+	if (ap->ap_flags & A_FREE_REPORT)
+		free(ap->ap_report);
+	if (ap->ap_flags & A_FREE_ADDHEADER)
+		free(ap->ap_addheader);
+	if (ap->ap_flags & A_FREE_ADDFOOTER)
+		free(ap->ap_addfooter);
+	if (ap->ap_flags & A_FREE_SUBJTAG)
+		free(ap->ap_subjtag);
+
+	if (ap->ap_flags & A_FLUSHADDR) {
+		struct tuple_fields tuple;
+
+		tuple.sa = sa;
+		tuple.salen = salen;
+		tuple.queueid = priv->priv_queueid;
+		tuple.acl_line = acl->a_line;
+
+		mg_tuple_remove(&tuple);
+		}
+
+	if (ap->ap_flags & A_NOLOG)
+		retval |= EXF_NOLOG;
+
+	if (conf.c_debug || conf.c_acldebug) {
+		char aclstr[HDRLEN + 1];
+
+		iptostring(sa, salen, addrstr, sizeof(addrstr));
+		mg_log(LOG_DEBUG, "Mail from=%s, rcpt=%s, addr=%s[%s] "
+		    "is matched by entry %s", priv->priv_from, 
+		    (priv->priv_cur_rcpt[0] != '\0') ? priv->priv_cur_rcpt : "(nil)",
+		    priv->priv_hostname, addrstr, 
+		    acl_entry(aclstr, HDRLEN, acl));
+	}
+
+	return retval;
+}
+
 int
 acl_filter(stage, ctx, priv)
 	acl_stage_t stage;
@@ -1952,6 +2359,11 @@ acl_filter(stage, ctx, priv)
 		if (acl->a_stage != stage)
 			continue;
 
+		/*
+		 * Keep track of current ACL for %cA and %ca format strings
+		 */
+		priv->priv_acl = acl;
+
 		retval = 0;
 		noretval = 0;
 		found = -1;
@@ -1969,6 +2381,9 @@ acl_filter(stage, ctx, priv)
 		ap.ap_msg = acl->a_msg;
 		ap.ap_report = acl->a_report;
 		ap.ap_addheader = acl->a_addheader;
+		ap.ap_addfooter = acl->a_addfooter;
+		ap.ap_subjtag = acl->a_subjtag;
+		ap.ap_maxpeek = acl->a_maxpeek;
 
 		/*
 		 * Free pointers to stored parenthesized substrings 
@@ -2010,136 +2425,17 @@ acl_filter(stage, ctx, priv)
 		}
 
 		/*
-		 * We catch the first ACL entry that matches, so exit
-		 * evaluation loop when found.
+		 * If the ACl matched, apply actions.
+		 * Then exit evaluation if it was not a continue ACL
 		 */
-		if (found != 0)
-			break;
+		if (found != 0) {
+			retval = acl_actions(priv, stage, acl, &ap, retval);
+			if (ap.ap_type != A_CONTINUE)
+				break;
+		}
 	}
 
-	if (acl) {
-		if (retval == 0)
-			retval = EXF_DEFAULT;
-		switch (ap.ap_type) {
-		case A_GREYLIST:
-			retval |= EXF_GREYLIST;
-			break;
-		case A_WHITELIST:
-			retval |= EXF_WHITELIST;
-			break;
-		case A_BLACKLIST:
-			retval |= EXF_BLACKLIST;
-			break;
-		default:
-			mg_log(LOG_ERR, "corrupted acl list");
-			exit(EX_SOFTWARE);
-			break;
-		}
-
-		priv->priv_sr.sr_acl_line = acl->a_line;
-
-		priv->priv_sr.sr_delay =
-		    (ap.ap_delay != -1) ? ap.ap_delay : conf.c_delay;
-		priv->priv_sr.sr_autowhite =
-		    (ap.ap_autowhite != -1) ? 
-		    ap.ap_autowhite : conf.c_autowhite_validity;
-		priv->priv_sr.sr_tarpit =
-		    (ap.ap_tarpit != -1) ? ap.ap_tarpit : conf.c_tarpit;
-		priv->priv_sr.sr_tarpit_scope =
-		    (ap.ap_tarpit_scope != -1) ?
-		    ap.ap_tarpit_scope : conf.c_tarpit_scope;
-		if (ap.ap_tarpitted > priv->priv_max_tarpitted)
-			priv->priv_max_tarpitted = ap.ap_tarpitted;
-		if (ap.ap_tarpitted > priv->priv_total_tarpitted)
-			priv->priv_total_tarpitted = ap.ap_tarpitted;
-
-		if (ap.ap_id) {
-			priv->priv_sr.sr_acl_id = strdup(ap.ap_id);
-			if (priv->priv_sr.sr_acl_id == NULL) { 
-				mg_log(LOG_ERR, "strdup failed: %s", 
-				    strerror(errno));
-				exit(EX_OSERR);
-			}
-		}
-		if (ap.ap_code) {
-			priv->priv_sr.sr_code = strdup(ap.ap_code);
-			if (priv->priv_sr.sr_code == NULL) { 
-				mg_log(LOG_ERR, "strdup failed: %s", 
-				    strerror(errno));
-				exit(EX_OSERR);
-			}
-		}
-		if (ap.ap_ecode) {
-			priv->priv_sr.sr_ecode = strdup(ap.ap_ecode);
-			if (priv->priv_sr.sr_ecode == NULL) {
-				mg_log(LOG_ERR, "strdup failed: %s", 
-				    strerror(errno));
-				exit(EX_OSERR);
-			}
-		}
-		if (ap.ap_msg) {
-			priv->priv_sr.sr_msg = strdup(ap.ap_msg);
-			if (priv->priv_sr.sr_msg == NULL) {
-				mg_log(LOG_ERR, "strdup failed");
-				exit(EX_OSERR);
-			}
-		}
-		if (ap.ap_report) {
-			priv->priv_sr.sr_report = strdup(ap.ap_report);
-			if (priv->priv_sr.sr_report == NULL) {
-				mg_log(LOG_ERR, "strdup failed");
-				exit(EX_OSERR);
-			}
-		}
-		if (ap.ap_addheader) {
-			priv->priv_sr.sr_addheader = strdup(ap.ap_addheader);
-			if (priv->priv_sr.sr_addheader == NULL) {
-				mg_log(LOG_ERR, "strdup failed");
-				exit(EX_OSERR);
-			}
-		}
-
-		priv->priv_sr.sr_nmatch = ap.ap_nmatch;
-		priv->priv_sr.sr_pmatch = ap.ap_pmatch;
-			
-		/* Free temporary memory if nescessary */
-		if (ap.ap_flags & A_FREE_ID)
-			free(ap.ap_id);
-		if (ap.ap_flags & A_FREE_CODE)
-			free(ap.ap_code);
-		if (ap.ap_flags & A_FREE_ECODE)
-			free(ap.ap_ecode);
-		if (ap.ap_flags & A_FREE_MSG)
-			free(ap.ap_msg);
-		if (ap.ap_flags & A_FREE_REPORT)
-			free(ap.ap_report);
-		if (ap.ap_flags & A_FREE_ADDHEADER)
-			free(ap.ap_addheader);
-
-		if (ap.ap_flags & A_FLUSHADDR) {
-			struct tuple_fields tuple;
-
-			tuple.sa = sa;
-			tuple.salen = salen;
-			tuple.queueid = queueid;
-			tuple.acl_line = acl->a_line;
-
-			mg_tuple_remove(&tuple);
-			}
-
-		if (ap.ap_flags & A_NOLOG)
-			retval |= EXF_NOLOG;
-
-		if (conf.c_debug || conf.c_acldebug) {
-			char aclstr[HDRLEN + 1];
-
-			iptostring(sa, salen, addrstr, sizeof(addrstr));
-			mg_log(LOG_DEBUG, "Mail from=%s, rcpt=%s, addr=%s[%s] "
-			    "is matched by entry %s", from, 
-			    (cur_rcpt != NULL) ? cur_rcpt : "(nil)",
-			    hostname, addrstr, acl_entry(aclstr, HDRLEN, acl));
-		}
-	} else {
+	if (acl == NULL) {
 		/*
 		 * No match: use the default action
 		 */
@@ -2197,7 +2493,7 @@ acl_filter(stage, ctx, priv)
 			ADD_REASON(whystr, tmpstr);
 		}
 		if (retval & EXF_RCPT) {
-			if (cur_rcpt != NULL) {
+			if (cur_rcpt[0] != '\0') {
 				snprintf(tmpstr, sizeof(tmpstr),
 				     "recipient %s is whitelisted", cur_rcpt);
 			} else {
@@ -2247,10 +2543,10 @@ acl_filter(stage, ctx, priv)
 			ADD_REASON(whystr, "this is the default action");
 		}
 		iptostring(sa, salen, addrstr, sizeof(addrstr));
-		aclstr = fstring_expand(priv, NULL, "%a");
+		aclstr = fstring_expand(priv, NULL, "%a", NULL);
 		snprintf(tmpstr, sizeof(tmpstr),
 		    "(from=%s, rcpt=%s, addr=%s[%s]) ACL %s", from, 
-		    (cur_rcpt != NULL) ? cur_rcpt : "(nil)",
+		    (cur_rcpt[0] != '\0') ? cur_rcpt : "(nil)",
 		    hostname, addrstr, aclstr);
 		free(aclstr);
 		ADD_REASON(whystr, tmpstr);
@@ -2373,6 +2669,9 @@ acl_entry(entrystr, len, acl)
 	case A_BLACKLIST:
 		mystrlcat(entrystr, "blacklist ", len);
 		break;
+	case A_CONTINUE:
+		mystrlcat(entrystr, "continue ", len);
+		break;
 	default:
 		mg_log(LOG_ERR, "corrupted acl list");
 		exit(EX_SOFTWARE);
@@ -2458,6 +2757,21 @@ acl_entry(entrystr, len, acl)
 		    "[addheader \"%s\"] ", acl->a_addheader);
 		mystrlcat(entrystr, tempstr, len);
 	}
+	if (acl->a_addfooter) {
+		snprintf(tempstr, sizeof(tempstr), 
+		    "[addfooter \"%s\"] ", acl->a_addfooter);
+		mystrlcat(entrystr, tempstr, len);
+	}
+	if (acl->a_subjtag) {
+		snprintf(tempstr, sizeof(tempstr), 
+		    "[subjtag \"%s\"] ", acl->a_subjtag);
+		mystrlcat(entrystr, tempstr, len);
+	}
+	if (acl->a_maxpeek) {
+		snprintf(tempstr, sizeof(tempstr), 
+		    "[maxpeek %d] ", acl->a_maxpeek);
+		mystrlcat(entrystr, tempstr, len);
+	}
 
 	if (def)
 		mystrlcat(entrystr, "default", len);
@@ -2719,7 +3033,7 @@ acl_add_report(report)
 }
 
 void 
- acl_add_addheader(hdr)
+acl_add_addheader(hdr)
 	char *hdr;
 {
 	if (gacl->a_addheader) {
@@ -2740,6 +3054,70 @@ void
 	return;
 }
 
+void 
+acl_add_addfooter(hdr)
+	char *hdr;
+{
+	if (gacl->a_addfooter) {
+		mg_log(LOG_ERR,
+		    "addfooter specified twice in ACL line %d", conf_line);
+		exit(EX_DATAERR);
+	}
+
+	if ((gacl->a_addfooter = strdup(hdr)) == NULL) {
+		mg_log(LOG_ERR,
+		    "malloc failed in ACL line %d", conf_line);
+		exit(EX_OSERR);
+	}
+		
+	if (conf.c_debug || conf.c_acldebug)
+		mg_log(LOG_DEBUG, "load acl addfooter \"%s\"", hdr);
+
+	return;
+}
+
+void 
+acl_add_subjtag(tag)
+	char *tag;
+{
+	if (gacl->a_subjtag) {
+		mg_log(LOG_ERR,
+		    "subjtag specified twice in ACL line %d", conf_line);
+		exit(EX_DATAERR);
+	}
+
+	if ((gacl->a_subjtag = strdup(tag)) == NULL) {
+		mg_log(LOG_ERR,
+		    "malloc failed in ACL line %d", conf_line);
+		exit(EX_OSERR);
+	}
+		
+	if (conf.c_debug || conf.c_acldebug)
+		mg_log(LOG_DEBUG, "load acl subjtag \"%s\"", tag);
+
+	return;
+}
+
+void
+acl_add_maxpeek(maxpeek)
+	int maxpeek;
+{
+	gacl->a_maxpeek = maxpeek;
+
+	if (conf.c_debug || conf.c_acldebug)
+		mg_log(LOG_DEBUG, "load acl maxpeek %d", maxpeek);
+
+	return;
+}
+
+void
+acl_maxpeek_fixup(maxpeek)
+	int maxpeek;
+{
+	gacl->a_maxpeek = maxpeek;
+	return;
+}
+
 int
 acl_modify_by_prop(key, value, ap)
 	char *key;
@@ -2762,6 +3140,8 @@ acl_modify_by_prop(key, value, ap)
 			ap->ap_type = A_BLACKLIST;
 		else if (strcasecmp(value, "whitelist") == 0)
 			ap->ap_type = A_WHITELIST;
+		else if (strcasecmp(value, "continue") == 0)
+			ap->ap_type = A_CONTINUE;
 		else
 			mg_log(LOG_WARNING, "ignored greylist-type \"%s\"",
 			    value);
@@ -2847,3 +3227,20 @@ out:
 
 	return 0;
 }
+
+int 
+acl_log_string(ad, stage, ap, priv)
+	acl_data_t *ad; 
+	acl_stage_t stage; 
+	struct acl_param *ap;
+	struct mlfi_priv *priv;
+{
+	char *string;
+
+	string = fstring_expand(priv, NULL, ad->string, NULL);
+	mg_log(LOG_INFO, string);	
+	free(string);
+
+	/* always match */
+	return 1;
+}
