@@ -1,7 +1,7 @@
-/* $Id: acl.h,v 1.44 2010/04/18 16:18:27 manu Exp $ */
+/* $Id: acl.h,v 1.54 2013/08/27 13:50:44 manu Exp $ */
 
 /*
- * Copyright (c) 2004-2007 Emmanuel Dreyfus
+ * Copyright (c) 2004-2012 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
 #define _ACL_H_
 
 #include "config.h"
-#ifdef HAVE_OLD_QUEUE_H
+#if defined(HAVE_OLD_QUEUE_H) || !defined(HAVE_SYS_QUEUE_H)
 #include "queue.h"
 #else 
 #include <sys/queue.h>
@@ -48,11 +48,11 @@
 #include <arpa/inet.h>
 #include <regex.h>
 
-typedef enum { A_GREYLIST, A_WHITELIST, A_BLACKLIST, } acl_type_t;
+typedef enum { A_GREYLIST, A_WHITELIST, A_BLACKLIST, A_CONTINUE, } acl_type_t;
 typedef enum { AS_NONE, AS_RCPT, AS_DATA, AS_ANY, } acl_stage_t;
 typedef enum { AT_NONE, AT_STRING, AT_REGEX, AT_NETBLOCK, AT_OPNUM, 
 	       AT_CLOCKSPEC, AT_DNSRBL, AT_MX, AT_URLCHECK, AT_MACRO, 
-	       AT_LIST, AT_PROP, AT_SPF, AT_DKIM, 
+	       AT_LIST, AT_PROP, AT_SPF, AT_DKIM, AT_NSUPDATE,
 	       AT_LDAPCHECK, AT_TIME, AT_RATELIMIT } acl_data_type_t;
 
 typedef enum {
@@ -61,6 +61,7 @@ typedef enum {
 	AC_EMAIL,
 	AC_REGEX,
 	AC_STRING,
+	AC_STRING_PROP,
 	AC_HELO,
 	AC_HELO_RE,
 	AC_HELO_LIST,
@@ -78,9 +79,11 @@ typedef enum {
 	AC_BODY,
 	AC_BODY_LIST,
 	AC_BODY_RE,
+	AC_BODY_PROP,
 	AC_HEADER,
 	AC_HEADER_LIST,
 	AC_HEADER_RE,
+	AC_HEADER_PROP,
 	AC_DNSRBL,
 	AC_DNSRBL_LIST,
 	AC_MX,
@@ -90,8 +93,11 @@ typedef enum {
 	AC_URLCHECK,
 	AC_URLCHECK_LIST,
 	AC_LDAPCHECK,
-	AC_PROP,
+	AC_PROP_STR,
+	AC_PROP_NUM,
+	AC_PROP_PROP,
 	AC_PROP_RE,
+	AC_HEADERPROP,
 	AC_AUTH,
 	AC_AUTH_RE,
 	AC_AUTH_LIST,
@@ -101,7 +107,9 @@ typedef enum {
 	AC_SPF,
 	AC_DKIM,
 	AC_MSGSIZE,
+	AC_MSGSIZE_PROP,
 	AC_RCPTCOUNT,
+	AC_RCPTCOUNT_PROP,
 	AC_CLOCKSPEC,
 	AC_CLOCKSPEC_LIST,
 	AC_GEOIP,
@@ -111,15 +119,34 @@ typedef enum {
 	AC_P0F_LIST,
 	AC_SA,
 	AC_SASCORE,
+	AC_SASCORE_PROP,
 	AC_TARPIT,
 	AC_RATELIMIT,
+	AC_NSUPDATE,
+	AC_EQSET,
+	AC_EQRSET,
+	AC_INCSET,
+	AC_INCRSET,
+	AC_DECSET,
+	AC_DECRSET,
+	AC_EQSETPROP,
+	AC_EQRSETPROP,
+	AC_INCSETPROP,
+	AC_INCRSETPROP,
+	AC_DECSETPROP,
+	AC_DECRSETPROP,
+	AC_LOG,
 } acl_clause_t;
 
+typedef union acl_data acl_data_t;
 struct acl_clause;
 struct acl_param;
 
 #include "pending.h"
 #include "milter-greylist.h"
+#ifdef USE_NSUPDATE
+#include "nsupdate.h"
+#endif
 
 #define ACL_WRLOCK WRLOCK(acl_lock) 
 #define ACL_RDLOCK RDLOCK(acl_lock) 
@@ -155,6 +182,9 @@ struct acl_param {
 	char *ap_msg;
 	char *ap_report;
 	char *ap_addheader;
+	char *ap_addfooter;
+	char *ap_subjtag;
+	int ap_maxpeek;
 	int ap_nmatch;
 	char **ap_pmatch;
 };
@@ -169,10 +199,12 @@ struct acl_param {
 #define A_DROP_ACL		0x040
 #define A_NOLOG			0x080
 #define A_FREE_ADDHEADER	0x100
+#define A_FREE_ADDFOOTER	0x200
+#define A_FREE_SUBJTAG		0x400
 
 struct all_list_entry;
 
-typedef union acl_data {
+union acl_data {
 	struct acl_netblock_data netblock;
 	char *string;
 	struct {
@@ -194,9 +226,9 @@ typedef union acl_data {
 #ifdef USE_LDAP
 	struct ldapcheck_entry *ldapcheck;
 #endif
-#if defined(USE_CURL) || defined(USE_LDAP)
 	struct prop_data *prop;
-#endif
+	struct acl_opnum_prop *aonp;
+	struct acl_prop_pop *apop;
 	struct acl_opnum_data opnum;
 	struct clockspec *clockspec;
 #if (defined(HAVE_SPF) || defined(HAVE_SPF_ALT) || \
@@ -208,7 +240,10 @@ typedef union acl_data {
 #endif
 	time_t time;
 	struct ratelimit_conf *ratelimit_conf;
-} acl_data_t;
+#ifdef USE_NSUPDATE
+	struct nsupdate_entry *nsupdate;
+#endif
+};
 
 struct acl_clause_rec {
 	acl_clause_t acr_type;
@@ -254,6 +289,9 @@ struct acl_entry {
 	char *a_msg;
 	char *a_report;
 	char *a_addheader;
+	char *a_addfooter;
+	char *a_subjtag;
+	int a_maxpeek;
 	TAILQ_ENTRY(acl_entry) a_list;
 };
 
@@ -279,6 +317,10 @@ void acl_add_ecode(char *);
 void acl_add_msg(char *);
 void acl_add_report(char *);
 void acl_add_addheader(char *);
+void acl_add_addfooter(char *);
+void acl_add_subjtag(char *);
+void acl_add_maxpeek(int);
+void acl_maxpeek_fixup(int);
 struct acl_entry *acl_register_entry_first(acl_stage_t, acl_type_t);
 struct acl_entry *acl_register_entry_last(acl_stage_t, acl_type_t);
 int acl_filter(acl_stage_t, SMFICTX *, struct mlfi_priv *);
