@@ -1,4 +1,4 @@
-/* $Id: dnsrbl.c,v 1.29 2010/03/13 07:00:12 manu Exp $ */
+/* $Id: dnsrbl.c,v 1.30 2013/01/19 16:01:15 manu Exp $ */
 
 /*
  * Copyright (c) 2006 Emmanuel Dreyfus
@@ -36,7 +36,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: dnsrbl.c,v 1.29 2010/03/13 07:00:12 manu Exp $");
+__RCSID("$Id: dnsrbl.c,v 1.30 2013/01/19 16:01:15 manu Exp $");
 #endif
 #endif
 
@@ -50,7 +50,7 @@ __RCSID("$Id: dnsrbl.c,v 1.29 2010/03/13 07:00:12 manu Exp $");
 #include <errno.h>
 #include <sysexits.h>
 
-#ifdef HAVE_OLD_QUEUE_H 
+#if defined(HAVE_OLD_QUEUE_H) || !defined(HAVE_SYS_QUEUE_H) 
 #include "queue.h"
 #else 
 #include <sys/queue.h>
