@@ -1,4 +1,4 @@
-/* $Id: geoip.c,v 1.4 2010/02/15 16:38:03 manu Exp $ */
+/* $Id: geoip.c,v 1.5 2013/03/23 01:43:01 manu Exp $ */
 
 /*
  * Copyright (c) 2007 Emmanuel Dreyfus
@@ -139,8 +139,7 @@ geoip_set_ccode(priv)
 	UNLOCK(geoip_lock);
 
 	if (priv->priv_ccode == NULL)
-		mg_log(LOG_DEBUG, "GeoIP failed to lookup ip '%s'", ipstr);
-
+		priv->priv_ccode = "ZZ";
 
 	return;
 }
