@@ -1,7 +1,7 @@
-/* $Id: milter-greylist.h,v 1.84 2010/07/12 01:38:14 manu Exp $ */
+/* $Id: milter-greylist.h,v 1.96 2014/02/07 13:46:46 manu Exp $ */
 
 /*
- * Copyright (c) 2004-2010 Emmanuel Dreyfus
+ * Copyright (c) 2004-2012 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -37,6 +37,10 @@
 #include <arpa/inet.h>
 #include <netdb.h>
 
+#include <libmilter/mfapi.h>
+#include "config.h"
+#include "dump.h"
+
 #ifdef USE_DKIM
 #ifdef HAVE_STDBOOL_H
 #include <stdbool.h>
@@ -44,15 +48,11 @@
 #include <dkim.h>
 #endif
 
-#include <libmilter/mfapi.h>
-#include "config.h"
-#include "dump.h"
-
 /* environment of Solaris workaround for stdio descriptor limitation */
 #include "fd_pool.h"
 
 #define NUMLEN 20
-#define QSTRLEN 1024
+#define QSTRLEN 4096
 #define REGEXLEN 1024
 #define HDRLEN 1024
 #define HEADERNAME "X-Greylist"
@@ -168,6 +168,8 @@ struct smtp_reply {
 	char *sr_report;
 	char *sr_report_x;
 	char *sr_addheader;
+	char *sr_addfooter;
+	char *sr_subjtag;
 	sfsistat sr_retcode;
 	int sr_nmatch;
 	char **sr_pmatch;
@@ -178,17 +180,13 @@ struct rcpt {
 	LIST_ENTRY(rcpt) r_list;
 };
 
-struct header {
-	char *h_line;
-	size_t h_len;
-	TAILQ_ENTRY(header) h_list;
+struct line {
+	char *l_line;
+	size_t l_len;
+	TAILQ_ENTRY(line) l_list;
 };
 
-struct body {
-	char *b_lines;
-	size_t b_len;
-	TAILQ_ENTRY(body) b_list;
-};
+TAILQ_HEAD(bh_line, line);
 
 struct mlfi_priv {
 	SMFICTX *priv_ctx;
@@ -198,14 +196,17 @@ struct mlfi_priv {
 	char priv_helo[ADDRLEN + 1];
 	char priv_from[ADDRLEN + 1];
 	LIST_HEAD(, rcpt) priv_rcpt;
-	char *priv_cur_rcpt;
+	char priv_cur_rcpt[ADDRLEN + 1];
 	int priv_rcptcount;
-	TAILQ_HEAD(, header) priv_header;
-	TAILQ_HEAD(, body) priv_body;
+	struct bh_line priv_header;
+	struct bh_line priv_body;
+	int priv_maxpeek;
 #ifdef USE_GEOIP
 	const char *priv_ccode;
 #endif
 	size_t priv_msgcount;
+	size_t priv_peekcount;
+	int priv_peekdone;
 	char *priv_buf;
 	size_t priv_buflen;
 	char *priv_queueid;
@@ -213,9 +214,8 @@ struct mlfi_priv {
 	struct smtp_reply priv_sr;
 	time_t priv_max_elapsed;
 	long long priv_last_whitelist;
-#if defined(USE_CURL) || defined(USE_LDAP)
 	LIST_HEAD(, prop) priv_prop;
-#endif
+	struct prop *priv_prop_match;
 #ifdef USE_DNSRBL
 	LIST_HEAD(, dnsrbl_list) priv_dnsrbl;
 #endif
@@ -232,6 +232,7 @@ struct mlfi_priv {
 #endif
 	time_t priv_max_tarpitted;
 	time_t priv_total_tarpitted;
+	struct acl_entry *priv_acl;	/* current ACL being evaluated */
 };
 
 sfsistat mlfi_connect(SMFICTX *, char *, _SOCK_ADDR *);
@@ -262,8 +263,9 @@ int main(int, char **);
 void mg_log(int, char *, ...);
 char *strncpy_rmsp(char *, char *, size_t);
 char *fstring_expand(struct mlfi_priv *, 
-    char *, const char *);
+    char *, const char *, char *(*)(char *));
 char *fstring_escape(char *);
+void mkparentdir(char *, mode_t);
 
 #ifdef HAVE_STRLCAT
 /* #include <string.h> */
