@@ -1,4 +1,4 @@
-/* $Id: stat.c,v 1.6 2010/06/16 01:30:30 manu Exp $ */
+/* $Id: stat.c,v 1.12 2013/01/19 16:01:15 manu Exp $ */
 
 /*
  * Copyright (c) 2007-2010 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: stat.c,v 1.6 2010/06/16 01:30:30 manu Exp $");
+__RCSID("$Id: stat.c,v 1.12 2013/01/19 16:01:15 manu Exp $");
 #endif
 #endif
 
@@ -46,7 +46,7 @@ __RCSID("$Id: stat.c,v 1.6 2010/06/16 01:30:30 manu Exp $");
 #include <ctype.h>
 #include <sysexits.h>
 
-#ifdef HAVE_OLD_QUEUE_H 
+#if defined(HAVE_OLD_QUEUE_H) || !defined(HAVE_SYS_QUEUE_H) 
 #include "queue.h"
 #else 
 #include <sys/queue.h>
@@ -171,14 +171,16 @@ mg_stat(priv, stat)
 	 */
 	priv->priv_sr.sr_retcode = stat; 
 
-	if (priv->priv_cur_rcpt == NULL) {
+	if (priv->priv_cur_rcpt[0] == '\0') {
 		LIST_FOREACH(rcpt, &priv->priv_rcpt, r_list) {
-			statlog = fstring_expand(priv, rcpt->r_addr, format);
+			statlog = fstring_expand(priv, rcpt->r_addr,
+						 format, NULL);
 			fprintf(outfp, "%s", statlog);
 			free(statlog);
 		}
 	} else {
-		statlog = fstring_expand(priv, priv->priv_cur_rcpt, format);
+		statlog = fstring_expand(priv, priv->priv_cur_rcpt,
+					 format, NULL);
 		fprintf(outfp, "%s", statlog);
 		free(statlog);
 	}
